"""Physical constants from NIST CODATA 2022 and PubChem. All values must have a reference to their source.

Developer Notes:
    - The constants are loaded from a CSV file and set as module attributes.
    - The `Constant` class is a subclass of `float` that holds additional metadata.
    - The `as_list` function returns a list of all defined constants.
    - The `show` function prints a human-readable table of all constants.
    - The `sources` list contains all sources for constants and conversion factors.
    - I considered subclassing dict for a ConstantsDict so that it has typical .keys(), .values(), and .items() methods.
        However, I decided against it because one looses IDE autocompletion for the constants as module-level attributes,
        which is more convenient for users.

References:
    - <https://physics.nist.gov/cuu/Constants/Table/allascii.txt>
"""

import csv
import inspect
import sys
from pathlib import Path

from .utils import to_table


class Constant(float):
    """A float subclass holding additional metadata."""

    unit: str
    source: str
    notes: str

    def __new__(cls, value, unit: str, source: str, notes: str):
        obj = float.__new__(cls, value)
        obj.unit = unit
        obj.source = source
        obj.notes = notes
        return obj

    def __repr__(self):
        return f"{float(self)}, unit={self.unit}, source={self.source} {'notes=' + self.notes if self.notes else ''}"


def _load_constants(csv_path: Path):
    """Load constants from CSV and set as module attributes."""
    current_module = sys.modules[__name__]
    with csv_path.open() as f:
        reader = csv.DictReader(f)
        for row in reader:
            name = row["name"]
            value = float(row["value"])
            unit = row["unit"]
            source = row["source"]
            notes = row["notes"]
            setattr(current_module, name, Constant(value, unit, source, notes))


# Load constants at import time
_load_constants(Path(__file__).parent / "data" / "constants.csv")


def as_list():
    """Return a list of all defined constants."""
    current_module = sys.modules[__name__]
    return [
        name
        for name, value in inspect.getmembers(current_module)
        if name.isupper()
        and not inspect.ismodule(value)
        and not inspect.isfunction(value)
    ]


def show():
    """Print a human-readable table of all constants."""
    current_module = sys.modules[__name__]
    constants_list = []
    for name in as_list():
        constant = getattr(current_module, name)
        constants_list.append(
            {
                "name": name,
                "value": float(constant),
                "unit": constant.unit,
                "source": constant.source,
            }
        )
    to_table(constants_list)


def _sources():
    """List unique sources for constants and conversion factors."""
    current_module = sys.modules[__name__]
    unique_sources = set()

    for name in as_list():
        source = getattr(current_module, name).source
        if not source.startswith("Derived"):
            unique_sources.add(source)

    return sorted(unique_sources)


sources = _sources()



import pytest

from qcconst import PeriodicTable
from qcconst import periodic_table as pt


def test_periodic_table_from_pubchem():
    pt = PeriodicTable.from_pubchem()
    assert pt.sources == [
        "https://pubchem.ncbi.nlm.nih.gov/periodic-table",
        "https://pubchem.ncbi.nlm.nih.gov/rest/pug/periodictable/CSV?response_type=save&response_basename=PubChemElements_all",
    ]


def test_group_period_lookup_params():
    with pytest.raises(AssertionError):
        pt.group(0)
    with pytest.raises(AssertionError):
        pt.group(19)
    with pytest.raises(AssertionError):
        pt.period(0)
    with pytest.raises(AssertionError):
        pt.period(8)


def test_group_and_period_data():
    # fmt: off
    # Assert that all groups and periods are correct
    assert [a.symbol for a in pt.group(1)] == ["H", "Li", "Na", "K", "Rb", "Cs", "Fr"]
    assert [a.symbol for a in pt.group(2)] == ["Be", "Mg", "Ca", "Sr", "Ba", "Ra"]
    assert [a.symbol for a in pt.group(3)] == ["Sc", "Y", "Lu", "Lr"]
    assert [a.symbol for a in pt.group(4)] == ["Ti", "Zr", "Hf", "Rf"]
    assert [a.symbol for a in pt.group(5)] == ["V", "Nb", "Ta", "Db"]
    assert [a.symbol for a in pt.group(6)] == ["Cr", "Mo", "W", "Sg"]
    assert [a.symbol for a in pt.group(7)] == ["Mn", "Tc", "Re", "Bh"]
    assert [a.symbol for a in pt.group(8)] == ["Fe", "Ru", "Os", "Hs"]
    assert [a.symbol for a in pt.group(9)] == ["Co", "Rh", "Ir", "Mt"]
    assert [a.symbol for a in pt.group(10)] == ["Ni", "Pd", "Pt", "Ds"]
    assert [a.symbol for a in pt.group(11)] == ["Cu", "Ag", "Au", "Rg"]
    assert [a.symbol for a in pt.group(12)] == ["Zn", "Cd", "Hg", "Cn"]
    assert [a.symbol for a in pt.group(13)] == ["B", "Al", "Ga", "In", "Tl", "Nh"]
    assert [a.symbol for a in pt.group(14)] == ["C", "Si", "Ge", "Sn", "Pb", "Fl"]
    assert [a.symbol for a in pt.group(15)] == ["N", "P", "As", "Sb", "Bi", "Mc"]
    assert [a.symbol for a in pt.group(16)] == ["O", "S", "Se", "Te", "Po", "Lv"]
    assert [a.symbol for a in pt.group(17)] == ["F", "Cl", "Br", "I", "At", "Ts"]
    assert [a.symbol for a in pt.group(18)] == ["He", "Ne", "Ar", "Kr", "Xe", "Rn", "Og"]  # noqa: E501
    assert [a.symbol for a in pt.period(1)] == ["H", "He"]
    assert [a.symbol for a in pt.period(2)] == ["Li", "Be", "B", "C", "N", "O", "F", "Ne"]  # noqa: E501
    assert [a.symbol for a in pt.period(3)] == ["Na", "Mg", "Al", "Si", "P", "S", "Cl", "Ar"]  # noqa: E501
    assert [a.symbol for a in pt.period(4)] == ["K", "Ca", "Sc", "Ti", "V", "Cr", "Mn", "Fe", "Co", "Ni", "Cu", "Zn", "Ga", "Ge", "As", "Se", "Br", "Kr"]  # noqa: E501
    assert [a.symbol for a in pt.period(5)] == ["Rb", "Sr", "Y", "Zr", "Nb", "Mo", "Tc", "Ru", "Rh", "Pd", "Ag", "Cd", "In", "Sn", "Sb", "Te", "I", "Xe"]  # noqa: E501 
    assert [a.symbol for a in pt.period(6)] == ["Cs", "Ba", "La", "Ce", "Pr", "Nd", "Pm", "Sm", "Eu", "Gd", "Tb", "Dy", "Ho", "Er", "Tm", "Yb", "Lu", "Hf", "Ta", "W", "Re", "Os", "Ir", "Pt", "Au", "Hg", "Tl", "Pb", "Bi", "Po", "At", "Rn"]  # noqa: E501 
    assert [a.symbol for a in pt.period(7)] == ["Fr", "Ra", "Ac", "Th", "Pa", "U", "Np", "Pu", "Am", "Cm", "Bk", "Cf", "Es", "Fm", "Md", "No", "Lr", "Rf", "Db", "Sg", "Bh", "Hs", "Mt", "Ds", "Rg", "Cn", "Nh", "Fl", "Mc", "Lv", "Ts", "Og"]  # noqa: E501
    # fmt: on


def test_number():
    assert pt.number(1).symbol == "H"
    assert pt.number(118).symbol == "Og"
    with pytest.raises(ValueError):
        pt.number(0)


def test_sources():
    assert isinstance(pt.sources, list)


def test_show():
    pt.show()



"""A Periodic Table from PubChem data. Other data may be used to instantiate a Periodic Table too.
All values must have a reference to their source.

References:
    - PubChem periodic table data: <https://pubchem.ncbi.nlm.nih.gov/periodic-table/>. Click "Download"
        to get the CSV file.
"""

import csv
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional

from .utils import to_table


@dataclass(frozen=True)
class Atom:
    """Atom data model."""

    symbol: str
    number: int
    name: str
    mass: float
    group: Optional[int]
    period: int
    block: str
    electron_config: str


class PeriodicTable:
    """Periodic table data model. Data source and URL should be set by class methods."""

    # Should be set by @classmethod
    sources: List[str]

    @classmethod
    def from_pubchem(cls):
        """Create a periodic table from PubChem data."""
        instance = cls()
        instance.sources = [
            "https://pubchem.ncbi.nlm.nih.gov/periodic-table",
            "https://pubchem.ncbi.nlm.nih.gov/rest/pug/periodictable/CSV?response_type=save&response_basename=PubChemElements_all",
        ]

        group_and_period_data = {}

        data_dir = Path(__file__).parent / "data"

        with open(data_dir / "group_period.csv") as file:
            reader = csv.DictReader(file)
            for row in reader:
                group_and_period_data[row["Symbol"]] = {
                    "group": int(row["Group"]) if row["Group"] else None,
                    "period": int(row["Period"]),
                }

        with open(data_dir / "pubchem.csv") as file:
            reader = csv.DictReader(file)
            for row in reader:
                symbol = row["Symbol"]
                atom = Atom(
                    symbol=symbol,
                    number=int(row["AtomicNumber"]),
                    name=row["Name"],
                    mass=float(row["AtomicMass"]),
                    group=group_and_period_data[symbol]["group"],
                    period=group_and_period_data[symbol]["period"],
                    block=row["GroupBlock"],
                    electron_config=row["ElectronConfiguration"],
                )
                setattr(instance, symbol, atom)
        return instance

    def group(self, group_number: int) -> list[Atom]:
        """Return all atoms in a group."""
        assert 1 <= group_number <= 18, "Group number must be between 1 and 18."
        return [
            atom
            for atom in self.__dict__.values()
            if isinstance(atom, Atom) and atom.group == group_number
        ]

    def period(self, period_number: int) -> list[Atom]:
        """Return all atoms in a period."""
        assert 1 <= period_number <= 7, "Period number must be between 1 and 7."
        return [
            atom
            for atom in self.__dict__.values()
            if isinstance(atom, Atom) and atom.period == period_number
        ]

    def number(self, number: int) -> Atom:
        """Return an atom by atomic number."""
        for atom in self.__dict__.values():
            if isinstance(atom, Atom) and atom.number == number:
                return atom
        raise ValueError(f"No atom with atomic number {number}.")

    def show(self):
        """Print all atoms in a human-readable format."""
        print(f"Sources: {', '.join(self.sources)}\n")
        to_table([val for val in self.__dict__.values() if isinstance(val, Atom)])


periodic_table = PeriodicTable.from_pubchem()
"""Periodic table data from PubChem.

    Attributes:
        data_source (str): Data source.
        data_url (str): Data URL.

    Example:
        ```python
        >>> from qcconst import periodic_table as pt
        
        >>> pt.Ni
        Atom(symbol='Ni', number=28, name='Nickel', mass=58.6934, group=10, period=4, block='d', electron_config='[Ar] 3d8 4s2')

        >>> pt.group(1)
        [Atom(symbol='H', number=1, name='Hydrogen', mass=1.00784, group=1, period=1, block='s', electron_config='1s1'),  # noqa: E501
        Atom(symbol='Li', number=3, name='Lithium', mass=6.94, group=1, period=2, block='s', electron_config='[He] 2s1'),  # noqa: E501
        Atom(symbol='Na', number=11, name='Sodium', mass=22.98976928, group=1, period=3, block='s', electron_config='[Ne] 3s1'),  # noqa: E501
        Atom(symbol='K', number=19, name='Potassium', mass=39.0983, group=1, period=4, block='s', electron_config='[Ar] 4s1'),  # noqa: E501
        Atom(symbol='Rb', number=37, name='Rubidium', mass=85.4678, group=1, period=5, block='s', electron_config='[Kr] 5s1'),  # noqa: E501
        Atom(symbol='Cs', number=55, name='Cesium', mass=132.90545196, group=1, period=6, block='s', electron_config='[Xe] 6s1'),  # noqa: E501
        Atom(symbol='Fr', number=87, name='Francium', mass=223.0, group=1, period=7, block='s', electron_config='[Rn] 7s1')]  # noqa: E501
        ```
"""



from qcconst import solvents


def test_solvents():
    # Assert that all solvents have a name, dielectric and sources
    for name in solvents.as_list():
        solvent = solvents[name]
        assert solvent.name
        assert solvent.dielectric
        assert isinstance(solvent.sources, list)
        assert solvent.sources


def test_show():
    solvents.show()



from qcconst import constants


def test_constants():
    # Assert that all constants have a unit and source
    for name in constants.as_list():
        constant = getattr(constants, name)
        assert constant  # Not None
        assert constant.unit
        assert constant.source


def test_sources():
    assert isinstance(constants.sources, list)


def test_show():
    constants.show()


def test_repr():
    constant = constants.HARTREE_TO_KCAL_PER_MOL
    assert isinstance(repr(constant), str)



"""Solvent data model.

References:
    - Dielectric constants: <https://depts.washington.edu/eooptic/linkfiles/dielectric_chart%5B1%5D.pdf>
"""

import csv
from dataclasses import dataclass
from pathlib import Path

from .utils import to_table


@dataclass(frozen=True)
class Solvent:
    """Solvent data model."""

    name: str
    dielectric: float
    sources: list[str]


class SolventDict(dict):
    """Custom dictionary for solvents with extra convenience methods."""

    def as_list(self):
        """Return a list of all solvents."""
        return list(self.keys())

    def show(self):
        """Print a human-readable table of solvents and their dielectric constants."""
        to_table(list(self.values()))


solvents = SolventDict()


def _load_solvents(csv_path: Path):
    """Load solvents from CSV into the SOLVENTS dictionary."""
    with csv_path.open() as f:
        reader = csv.DictReader(f)
        for row in reader:
            name = row["name"]
            dielectric = float(row["dielectric"])
            sources = row["sources"].split(",")
            solvents[name] = Solvent(name=name, dielectric=dielectric, sources=sources)


# Load on import
_load_solvents(Path(__file__).parent / "data" / "solvents.csv")



from dataclasses import fields, is_dataclass
from typing import Optional

from tabulate import tabulate


def to_table(objects, headers: Optional[list[str]] = None):
    """Convert a list of dataclass objects into a nicely formatted table."""
    if not objects:
        print("No data to display.")
        return

    first_obj = objects[0]

    if isinstance(first_obj, dict):
        headers = headers or list(first_obj.keys())
        rows = [[obj.get(h, "") for h in headers] for obj in objects]
    elif is_dataclass(first_obj):
        headers = headers or [field.name for field in fields(first_obj)]
        rows = [[getattr(obj, h, "") for h in headers] for obj in objects]
    else:  # Regular Python objects
        headers = headers or list(vars(first_obj).keys())
        rows = [[getattr(obj, h, "") for h in headers] for obj in objects]

    print(tabulate(rows, headers=headers))
